var gotWarningsFromAllProjects=false;
var allDisplayedProjects=null;

// The number of projects that got into the current selection (for which
// warnings were last fetched).
// A project may fail to get in the selection if it doesn't have the current
// branch, or if it doesn't have snapshots indicated by inspection/previous
// branches.
var selectedProjectNum = 0;

// List of all branches (appearing for at least one project).
var allBranches = [];

var firstWarningsLoading = true;

var currentSvaceProjectIndex = 0;
var updatingAllProjects = false;

/**
 * This objects controls state of all branches in all projects
 */
var LFAcontroller = null;
var oldLFABranch=null;

//Generate the list allBranches of all branches in all projects
function fetchBranchesLFA() {
	var allProjects = [];
	for (var i=0; i<allDisplayedProjects.length; i++) {
		var proj = allDisplayedProjects[i];
		var controller = makeBranchController(proj, null, null, null, null);
		proj.ui.controller = controller;
		allProjects.push(proj);
	}
	var receiveGlobalController = function(controller){
		LFAcontroller = controller;
	}
	wrapControllers(allProjects, $("#branchContainer"), receiveGlobalController, function(){
		drop(SVACE_CLEANUP_SNAPSHOT);
		updateWarnings();
	});
}

/**
 * This function iterates over a given set of projects with given function.
 * We cannot use simple cycle since given function may start a background thread, and we do not want
 * such methods to interfere. We feed callback to a launched function to call after it has finished.
 * @param projects
 * @param step
 * @param end
 */
function processAllProjects(projects, step, end){
	var i = -1;
	processNextProject = function(){
		i++;
		if (i<projects.length){
			var project = projects[i];
			step(project, processNextProject);
		} else {
			end();
		}
	}
	processNextProject();
}

//Fetch warnings from all projects
function fetchWarningsLFA() {
	//we should fetch data only for projects having current branch
	var projects = LFAcontroller.getActiveProjects();
	var processed = 0;
	processAllProjects(projects, function(project, callback){
		currentSvaceProject = project;
		processed++;
		console.log("LFA: "+processed+"/"+projects.length);
		$('#warningsLoadingCounter').text(" project "+processed+"/"+projects.length);
		fetchWarnings({
			clearContainer: false,
			errorFunction: function (data, textstatus, error) {
				setError("Fatal: Couldn't fetch warnings for project " + project.project_name + "; " + error);
			},
			completeFunction: callback
		});
	}, fetchWarningsLFAEnd);
}

//What to do when warnings in all projects are fetched
function fetchWarningsLFAEnd() {
	fetching = false;
	gotWarningsFromAllProjects = true;
	gotWarnings = true;
	$('#warningsLoading').hide();
	//call for update immediately after fetch to fetch recent changes
	updateWarnings();
}

//To update projects, we have to check out which projects have been really updated (it's a cheap call),
//and then only fetch updates for updated projects.
function updateWarningsLFA() {
	//we should fetch data only for projects having current branch
	var projects = LFAcontroller.getActiveProjects();
	if (projects.length==0){
		updateWarningsLFAEnd();
		return;
	}
	var updated_projects=[];
	$.ajax( {
		url:'/history/api?act=get_modified_projects',
		dataType: 'json',
		type: 'GET',
		async: false,
		cache: false,
		success: function (data, textStatus) {
			//data is already json
			if (data.status != "OK") {
				setError("Error:"+data.error);
			} else {
				updated_projects = data.updated_projects;
			}
		},
		error: ajaxError
	});
	
	var projectsToFetch = [];
	for (var i = 0; i< projects.length; i++){
		var project = projects[i];
		if (updated_projects.indexOf(project.project_id)>=0){
			projectsToFetch.push(project);
		}
	}
	
	processAllProjects(projectsToFetch, function(project, callback){
		fetchUpdatesImpl(project);
		//No let's wait for update to complete
		var wait = function(){
			if (true === project.is_fetching_updates){
				setTimeout(wait, 1);
				return;
			}
			callback();
		}
		wait();
	}, updateWarningsLFAEnd);
}

function updateWarningsLFAEnd() {
	updatingAllProjects = false;
}

function oneSDInterval(marked, total) {
    // Estimate success rate
    var pEst = (marked+1) / (total+2);
    // Estimate standard deviation of our success rate estimate
    var sdEst = Math.sqrt(pEst*(1-pEst)/(total+1));
    // Lower and upper bounds of the interval
    var l = Math.max(pEst-sdEst, 0.0);
    var u = Math.min(pEst+sdEst, 1.0);
    return [l,u];
}

function percentString(marked, total) {
    var lu = oneSDInterval(marked, total);
    var l = Math.round(100.0*lu[0]);
    var u = Math.round(100.0*lu[1]);
    return l + "-" + u + "%";
}

function displaySorted(warnIndex, totalNum) {
    var totalNum = 0;
    var warnArray = [];
    for (var warnName in warnIndex) {
        var warnNum = warnIndex[warnName];
        totalNum += warnNum;
        warnArray.push([warnName, warnNum]);
    }

    warnArray.sort(function(a,b) {
        // Warning type name is at index 0, number of issues is at index 1
        if (b[1] !== a[1])
          return b[1] - a[1];
        if (b[0] > a[0])
            return -1;
        if (b[0] < a[0])
            return 1;
        return 0;
    });

    var resText = "";
    for (var i=0; i<warnArray.length; i++) {
        var warnName = warnArray[i][0];
        var warnNum = warnArray[i][1];
        resText += "* " + warnName + " (" + warnNum + ", " + percentString(warnNum, totalNum) + ")\n";
    }
    return resText;
}

function displayAllIssues(allIssues, falseIssues, trueIssues) {
    var warnArray = [];
    for (var warn in allIssues)
        warnArray.push([warn, allIssues[warn]]);

    warnArray.sort(function(a,b) {
        // Name is at index 0, number of issues is at index 1
        var percentA = (a[0] in trueIssues) ? trueIssues[a[0]] / a[1] : 0.0;
        var percentB = (b[0] in trueIssues) ? trueIssues[b[0]] / b[1] : 0.0;
        if (percentB !== percentA)
            return percentB - percentA;
        if (b[0] > a[0])
            return -1;
        if (b[0] < a[0])
            return 1;
        return 0;
    });

    var resText = "";
    for (var i=0; i<warnArray.length; i++) {
        var warnName = warnArray[i][0];
        var warnNum = warnArray[i][1];
        var trueNum = (warnName in trueIssues) ? trueIssues[warnName] : 0;
        var falseNum =  (warnName in falseIssues) ? falseIssues[warnName] : 0;
        var percentOfTrue = 100.0 * trueNum / warnNum;
        resText += "* " + warnName + " (" + warnNum + ") [" + trueNum + "/" + (trueNum + falseNum) + "] " + percentOfTrue.toFixed(1) + "%" + "\n";
    }
    return resText; 
}

/* Temporary stub: displayForSelectionTool is a flag which is defined and is true when we shouldn't use global variables.
   Check out FIXME lines at this method and fix em. */
function countAndDisplayStats(Warnings, displayForSelectionTool) {
	var allCheckerIssues = {};
	var falseCheckerIssues = {};
	var trueCheckerIssues = {};

	var allProjectIssues = {};
	var falseProjectIssues = {};
	var trueProjectIssues = {};

	var uninspectedCheckers = [];

	var falseNum = 0;
	var trueNum = 0;
	var scopeNum = 0;
	var uninspectedNum = 0;
	for (var warn_id in Warnings) {
		var Warning = Warnings[warn_id];
		var checkerName = getCheckerName(Warning.checker_id);
		var projectName = Warning.project.project_name;

		if(Warning.attributes.classification_id === 2)
			scopeNum++;

		if ((Warning.attributes.classification_id === 1) || (Warning.attributes.classification_id === 2)) {
			// true positive
			(checkerName in trueCheckerIssues) ? trueCheckerIssues[checkerName]++ : trueCheckerIssues[checkerName] = 1;
      (projectName in trueProjectIssues) ? trueProjectIssues[projectName]++ : trueProjectIssues[projectName] = 1;
			trueNum++;
		} else if ((Warning.attributes.classification_id === 3) || (Warning.attributes.classification_id === 4)) {
			// false positive
			(checkerName in falseCheckerIssues) ? falseCheckerIssues[checkerName]++ : falseCheckerIssues[checkerName] = 1;
			(projectName in falseProjectIssues) ? falseProjectIssues[projectName]++ : falseProjectIssues[projectName] = 1;
			falseNum++;
		} else {
			if (uninspectedCheckers.indexOf(checkerName)==-1)
				uninspectedCheckers.push(checkerName);
			uninspectedNum++;
		}

    (checkerName in allCheckerIssues) ? allCheckerIssues[checkerName]++ : allCheckerIssues[checkerName] = 1;
    (projectName in allProjectIssues) ? allProjectIssues[projectName]++ : allProjectIssues[projectName] = 1;
	}

	var statsText = "";
	try {

	if (uninspectedNum > 0)
		statsText += "Selected issues: " + (trueNum + falseNum + uninspectedNum) + " (not all of these were evaluated)\n";
	statsText += "Inspected issues: " + trueNum + "/" + (trueNum+falseNum) + " (" + percentString(trueNum, trueNum+falseNum) + ") true positives,\n";
	statsText += "including " + scopeNum + " 'won't fix' issues\n";

	if(uninspectedNum > 0) {
		statsText += "\n";
		statsText += "Uninspected checkers:\n";
	
		uninspectedCheckers.forEach(function(checker) {
  			statsText += checker + "\n";
		});
	}

	if (falseNum > 0) {
		statsText += "\n";
		statsText += "False issues by checker:\n";
		statsText += displaySorted(falseCheckerIssues);
	}

	if (trueNum > 0) {
    	statsText += "\n";
		statsText += "True issues by checker:\n";
		statsText += displaySorted(trueCheckerIssues);
	}
  if (trueNum + falseNum > 0) {
    statsText += "\n";
    statsText += "Issues by checker:\n";
    statsText += "format: * TYPE (ALL) [True/Inspected] and % of True among ALL\n";
    statsText += displayAllIssues(allCheckerIssues, falseCheckerIssues, trueCheckerIssues);

    statsText += "\n";
    statsText += "Issues by project:\n";
    statsText += "format: * PROJECT (ALL) [True/Inspected] and % of True among ALL\n";
    statsText += displayAllIssues(allProjectIssues, falseProjectIssues, trueProjectIssues);
  }

	} catch(err) {
		statsText += "\n";
		statsText += "*** Exception: " + err.message;
	}

    $('#selectionToolStatsArea').val(statsText);
}


function fixLFA(){
	clearTimeout(timeout);
    $.xhrPool.abortAll();
	for (project_idx=0; project_idx<allDisplayedProjects.length; project_idx++){
		allDisplayedProjects[project_idx].is_fetching_updates = false;
	}
	updateWarnings();
}

function svaceAssignRandomOwners()
{
  var WarningObjects = WarningObjectsWrap.WarningObjects; 
  var skipAssigned=!($('#reassign').is(':checked'));
  var mode= $("#assignMode").val();
  function updateStats(assignmentStats){
	  var prefix = "";
	  if (assignmentStats.processed!=assignmentStats.total){
		  prefix="Assignment in progress...<br/>";
	  }
	  $("#assignmentProgress").html(prefix+"Actually assigned and skipped: "+assignmentStats.processed+"/"+assignmentStats.total+"; "+assignmentStats.failed+" failed");
  }
  
  function assignWarningToUser(warning_id, user, stats) {
	  var proj = WarningObjects[warning_id].project;
	  $.ajax({
      url : '/history/api?act=put_attribute&warning_id='
          + warning_id + getBranchParameter(proj) + getProjectParameter(proj),
      type : 'POST',
      data : {user_owner: user},
      dataType : 'json',
      cache : false,
      success: function(){
    	  stats.processed++;
    	  updateStats(stats);
    	  },
      error : function(data, textstatus, error){
    	  stats.processed++;
    	  stats.failed++;
    	  updateStats(stats);
    	  ajaxError(data, textstatus, error);
      }
    });
  }
  report = "";
  users = $('#randomOwnersField').val().trim().split(/\s+/);
  var assignmentStats= {};
  assignmentStats.total=0;
  assignmentStats.processed=0;
  assignmentStats.failed=0;
  if(users.length === 0 || (users.length === 1 && users[0] === ""))
  {
    report = "No users to assign"
  }
  else
  {
	var target = null;
    target = DisplayedWarnings;
	
	var targetArray=[];
	{
	    Object.keys(target).forEach(function(warning_id){
	        warning = WarningObjects[warning_id];
	        targetArray.push(warning);
	    });
		if (mode=='random'){
			function shuffle(array) {
				  var currentIndex = array.length, temporaryValue, randomIndex ;

				  // While there remain elements to shuffle...
				  while (0 !== currentIndex) {

				    // Pick a remaining element...
				    randomIndex = Math.floor(Math.random() * currentIndex);
				    currentIndex -= 1;

				    // And swap it with the current element.
				    temporaryValue = array[currentIndex];
				    array[currentIndex] = array[randomIndex];
				    array[randomIndex] = temporaryValue;
				  }

				  return array;
				}
			shuffle(targetArray);
		} else if (mode=='project'){
			targetArray.sort(function(a,b){
				return a.project.project_id-b.project.project_id;
			});
		} else {
			targetArray.sort(function(a,b){
				return a.checker_id-b.checker_id;
			});
		}
	}

	{
		var usersReport={};
		users.forEach(function(user){
			usersReport[user]=0;
		});
	}
	var assigner;
	{
		var warningsNum=targetArray.length;
		var warningsPerUser=Math.ceil(warningsNum*1.0/users.length);
		var userIDX=0;
		var currentUserDone = 0;
		var usersWithMoreWarns = users.length;
		if (warningsPerUser*users.length>warningsNum){
			var usersWithLessWarns = warningsPerUser*users.length-warningsNum;
			usersWithMoreWarns-=usersWithLessWarns;
		}
		function fairNextChoice(arr) {
			  if (currentUserDone==warningsPerUser){
				  currentUserDone=0;
				  userIDX++;
				  if (userIDX==usersWithMoreWarns){
					  warningsPerUser--;
				  }
			  }
			  currentUserDone++;
			  return arr[userIDX];
		    }
		function randomChoice(arr) {
			return arr[Math.floor(arr.length * Math.random())];
		}
		assigner = fairNextChoice;
	}

  
    assigned = {}
    skipped = []
    
    
    targetArray.forEach(function(warning){
      assignmentStats.total++;
      if(warning.attributes.user_attributes.user_owner&&skipAssigned)
      {
        skipped.push(warning);
        assignmentStats.processed++;
      }
      else
      {
        user = assigner(users);
        var data = new Object();
        data.user=user;
        data.warning=warning;
        assigned[warning.warning_id] = data;
        assignWarningToUser(warning.warning_id, user, assignmentStats);
        usersReport[user]++;
      }
    });
    
    function processWarning(warning){
    	return "["+getCheckerName(warning.checker_id)+"]"+warning.warning_id.substring(0,6)+"@"+warning.project.project_name;
    }
      
      skipped.forEach(function(warning)
      {
        report += "Skipped " + processWarning(warning) + "\n";
      })
      $.each(assigned, function(warning_id, data)
      {
        report += "Assigned to " + data.user + " "+processWarning(data.warning) + "\n";
      })
      report+="\n";
      users.forEach(function(user){
    	  report+="User "+user+": "+usersReport[user]+" warnings assigned\n";
      });
  }
  updateStats(assignmentStats);
  $("#svaceAssignRandomOwnersReport").val(report);
  $("#svaceAssignRandomOwnersReport").show();

  updateWarnings();
}
