String.prototype.splice = function( idx, rem, s ) {
    return (this.slice(0,idx) + s + this.slice(idx + Math.abs(rem)));
};

var lastLocation = null;
var lastProjectData = null;
var lastSnapshotData = null;

function loadSource(location, projectData, snapshotData, force) {
	if (typeof force == 'undefined'){
		force = false;
	}
	if (location==null || projectData==null || snapshotData==null) {
		document.getElementById('sourceCodeC').innerHTML="";
		$("#usagesForm").dialog("close");
		lastLocation = null;
		lastProjectData = null;
		lastSnapshotData = null;
		return;
	}
	var hash = location.hash;
	var path = location.file_path;
	if (hash==null) {
		setError("Svace doesn't have this source file");
		return;
	}
	var line = location.line;
	if ((!force)&&(lastLocation!=null)&&(location.hash==lastLocation.hash)&&
			(
					(location.snapshot_override==lastLocation.snapshot_override)||
					((location.snapshot_override==undefined) && (lastLocation.snapshot_override==snapshotData))||
					((lastLocation.snapshot_override==undefined) && (location.snapshot_override==snapshotData))
					)
			) {
		highlightLine(line);
		return;
	}
	document.getElementById('sourceCodeC').innerHTML="Loading "+location.file_path+"...";
	$.ajax( {
		url:'/history/api?act=source&hash='+hash+projectData,
		dataType: 'text',
		cache: true,
		success: function (data, textStatus) {
			lastLocation=location;
			lastProjectData=projectData;
			lastSnapshotData=snapshotData;
			updateTitle();
			if (isSHLUsed()) {
				document.getElementById('sourceCodeC').innerHTML="<script id='sourceCode' type='syntaxhighlighter' class='brush: cpp'></scr"+"ipt>";
				document.getElementById('sourceCode').innerHTML = '<![CDATA['+data+']]>';
				SyntaxHighlighter.highlight();
			} else {
				makeNavigableCode(document.getElementById('sourceCodeC'), data, path, location.snapshot_override, lastProjectData, lastSnapshotData);
			}
			if ((location.snapshot_override!=undefined)&&(location.snapshot_override!=lastSnapshotData))
				setError("Displaying source code, not related to the selected build!");
			highlightLine(line);
		},
		error: ajaxError
	});
}

function isSHLUsed(){
	return isOldRenderer;
}

function highlightLine(line) {
	if (isSHLUsed()){
		var HLClassName = 'highlightedline';
		//first let's remove old highlights...
		$('.'+HLClassName).removeClass(HLClassName);

		var parentID = "sourceCodeC";
		var lineEls = document.getElementsByClassName("number"+line);
		scrolled = false;


		for (var i = lineEls.length-1; i>=0; i--) {
			elt = lineEls[i];
			elt.className=HLClassName+' '+elt.className;
			if (!scrolled) {
				var scroll = -elt.style.height;
				while(elt.getAttribute('id')!=parentID) {
					scroll += elt.offsetTop;
					elt=elt.parentNode;
				}
				var visibleHeight = elt.clientHeight;
				scroll -= visibleHeight/2;
				$('#'+parentID).scrollTop(scroll);
				scrolled = true;
			}
		}
	} else {
		$(".highlightedline").removeClass("highlightedline");
		hlobj=$("#line_"+line);
		hlobj.addClass("highlightedline");
		linetop=hlobj.offset().top-$("#line_1").offset().top;
		remainingtop=$('#sourceCodeC').height()/2;
		$('#sourceCodeC').animate({
			scrollTop: (linetop-remainingtop)
		}, 50);
	}
}

function makeNavigableCode(element, code, path, dxrOverride, projectData, snapshotDataBuf) {
	renderSuccess = false;
	snapshotData = null;
	if (dxrOverride == undefined)
		snapshotData = snapshotDataBuf;
	else
		snapshotData = dxrOverride;
	$.ajax( {
		url:'/history/api?act=get_tokens&file='+encodeURIComponent(path)+'&snapshot_id='+snapshotData+projectData,
		dataType: 'json',
		cache: true,
		async: false,
		success: function (data, textStatus) {
			renderSuccess = true;
			drawNavigableCode(element, code, path, data.tokens, snapshotData, projectData);
		},
		error: ajaxError
	});
	if (!renderSuccess)
		drawNavigableCode(element, code, path, [], snapshotData, projectData);

}

function escapeHTML(what){
	return what.replace(/&/g, "&amp;").replace(/</g, "&lt;").replace(/>/g, "&gt;");
}

function getTokenType(type) {
	switch (type) {
	case 0:
		return "macro";
	case 1:
		return "variable";
	case 2:
		return "function";
	case 3:
		return "type";
	default:
		return "???";
	}
}

function drawNavigableCode(element, code, path, tokens, snapshotData, projectData) {
	var html = "";
	var codelines = code.split("\n");
	
	var warnsInFile={};
	if (WarningsByFiles[path]!=undefined){
		for (var i=0; i<WarningsByFiles[path].length; i++) {
			warnInFile = WarningsByFiles[path][i];
			if (warnsInFile[warnInFile.location.line]==undefined)
				warnsInFile[warnInFile.location.line]=[];
			warnsInFile[warnInFile.location.line].push(warnInFile);
		}
	}
		
	var offsets = Array.apply(null, new Array(codelines.length)).map(Number.prototype.valueOf,0);
	var tokensOnLine = [];//i dunno JS, so it's a linked list of tokens for each source line
	for (var idx=0; idx<tokens.length; idx++){
		token = tokens[idx];
		list = tokensOnLine[token.location.line];
		if (list == undefined){
			list = new Object();
			tokensOnLine[token.location.line] = list;
		} else {
			while (list.next != undefined)
				list = list.next;
			list.next = new Object();
			list = list.next;
		}
		list.data = token;
	}
	
/*	for (var idx=0; idx<tokens.length; idx++){
		token = tokens[idx];
		line=codelines[token.location.line-1];
		start = token.offset+offsets[token.location.line];
		end   = start+token.length;
		oldlength=line.length;
		vr = line.substring(start, end);
		line = line.splice(start, token.length, "<a id='syntax_token_"+idx+"' href='javascript:;' onclick='drawSourceTooltip("+idx+","+token.token_group+");return false;' title='"+vr+"'>"+vr+"</a>");
		offsets[token.location.line]+=line.length-oldlength;
		codelines[token.location.line-1] = line;
	}*/
	
	maxNumberLength = codelines.length.toString().length;
	var tabs = Array(parseInt(spacesPerTab)+1).join("&nbsp;");
	for (var lidx=0; lidx<codelines.length; lidx++){
		space=maxNumberLength-(lidx+1).toString().length;
		if (space!=0)
			space = Array(space+1).join(" ");
		else
			space="";


		ltokenlist=tokensOnLine[lidx+1];//since token's line starts from 1, and line's array first element is [0]
		oldline=codelines[lidx];
		newline="";
		if (ltokenlist!=undefined){
			ltokens=[];
			while (ltokenlist!=undefined){
				ltokens[ltokens.length]=ltokenlist.data;
				ltokenlist=ltokenlist.next;
			}
			ltokens=ltokens.sort(function(a,b){return a.offset-b.offset;});
			newline="";
			prevOffset = 0;
			for (tidx=0; tidx<ltokens.length; tidx++){
				token=ltokens[tidx];
				if (token.offset+token.length>=prevOffset){ //otherwise it's fully under previous token. sad but such life in source code.
					realoffset = Math.max(token.offset, prevOffset);
					broken = (realoffset!=token.offset);
					newline += escapeHTML(oldline.substring(prevOffset,realoffset))
					ttvalue = escapeHTML(oldline.substring(token.offset, token.offset+token.length));
					value = escapeHTML(oldline.substring(realoffset,  token.offset+token.length));
					if (broken) {
						ttvalue = "Possibly misplaced entry caused by macro expansion: "+ttvalue;
						value = "<font color='#a000c0'>"+value+"</font>";
					}
					title = ttvalue+" ["+getTokenType(token.type)+"] ";
					newline+="<a class='"+token.token_group+"' id='syntax_token_"+lidx+"_"+tidx+"' href='javascript:;' onclick='drawSourceTooltip(\""+lidx+"_"+tidx+"\","+token.token_group+",\""+snapshotData+"\", \""+projectData+"\");return false;' title='"+title+"'>"+value+"</a>"
					prevOffset = token.offset+token.length;
				}
			}
			newline+=escapeHTML(oldline.substring(prevOffset, oldline.length));
		} else {
			newline=escapeHTML(oldline);
		}
		newline=newline.replace(/\t/g, tabs);

		lnumber=space+(lidx+1)+"|";
		if (warnsInFile[lidx+1]!=undefined){
			// onclick='displayWarningByIDAndClass(\""+warnsInFile[lidx+1].warning_id+"\", \""+warnsInFile[lidx+1].checker_id+"\")'
			lnumber="<div id='num_"+(lidx+1)+"' class='warnedline' title='"+warnsInFile[lidx+1].length+" warning(s) on this line (Click to list)'>"+lnumber+"</div>";
		}
		codelines[lidx]="<div style='display:block' id='line_"+(lidx+1)+"'>"+lnumber+newline+"</div>";
	}
	html="<pre>"+codelines.join("")+"</pre>";
	element.innerHTML = html;
	drawWarnsOnLineTooltip(warnsInFile)
}

function drawWarnsOnLineTooltip(warnsByLine){
	for (line in warnsByLine) {
		if (warnsByLine.hasOwnProperty(line)) {
		    $('#num_'+line).click(
		    	(function(lline){
		    		return function(){
		    			$('div.qtip:visible').remove();
		    			var content="";
		    			wsil = warnsByLine[lline];
		    			for (wil=0; wil<wsil.length; wil++){
		    				warn = wsil[wil];
		    				warning_id=warn.warning_id.toString();
		    				if (warning_id.length>6)
		    					warning_id=warning_id.substring(0,6)+"...";
		    				content+="<a href='javascript:;' onclick='$(\".qtip\").qtip(\"hide\");displayWarningByIDAndClass(\""+warn.warning_id+"\",\""+warn.checker_id+"\")'>"
		    				+"["+warning_id+"]"+getCheckerName(warn.checker_id)
		    				+"</a><br/>";
		    				if (wil==5){
		    					content+="And more";
		    					break;
		    				}
		    			}
		    			$('#num_'+lline).qtip({
		    				suppress: true,
		    				id:'numbers',
		    				show: {
		    					ready: true,
		    					solo: true,
		    				},
		    				hide : {
		    					fixed : true,
		    					event : '',
		    				},
		    				content : {
		    					title : "Warnings on line " + lline,
		    					text : content,
		    					button: true,
		    				},
		    				position : {
		    					container : $('#sourceCodeC'),
		    				},
		    			});
		    		};})(line))
		}
	}
}

function drawSourceTooltip(idx, group, snapshotData, projectData){
	//first click highlights; second shows tooltip
	if (!$('#syntax_token_'+idx).hasClass("SYNTAX_TOKEN_HGHL")){
		$(".SYNTAX_TOKEN_HGHL").removeClass("SYNTAX_TOKEN_HGHL");
		$("."+group).addClass("SYNTAX_TOKEN_HGHL");
		$('div.qtip:visible').remove();
	} else {
		$('#syntax_token_'+idx).qtip({
			suppress: true,
			id: idx,
			show: {
				ready: true,
				solo: true,
			},
			hide:{
				fixed:true,
				event:'',
			},
			content: {
				title:"Navigation",
				text:
					"<a href=javascript:; onclick=\"disableTip('"+idx+"');findDe("+group+",'def','"+snapshotData+"', '"+projectData+"');\">Find definition</a><br/>"+
					"<a href=javascript:; onclick=\"disableTip('"+idx+"');findDe("+group+",'decl','"+snapshotData+"', '"+projectData+"');\">Find declaration</a><br/>"+
					"<a href=javascript:; onclick=\"disableTip('"+idx+"');findUsage("+group+",'"+snapshotData+"', '"+projectData+"');\">Show usage</a><br/>",
					button: true,
			},
			position: {
				container:$('#sourceCodeC'),
			},
		});
	}
}

function disableTip(idx){
    $('#syntax_token_'+idx).qtip('destroy');
}

function findDe(group, kind, snapshotData, projectData){
	$.ajax( {
		url:'/history/api?act=get_ref_tokens&what='+kind+'&token_group='+group+'&snapshot_id='+snapshotData+projectData,
		dataType: 'json',
		cache: false,
		success: function (data, textStatus) {
			loadSource(data.tokens[0].location, projectData, snapshotData);
		},
		error: ajaxError
	});
}

function findUsage(group, snapshotData, projectData){
	$('#usagesContainer').html("Loading usage data...");
	$('#usagesForm').dialog("open");
	$.ajax( {
		url:'/history/api?act=get_ref_tokens&what=usage&token_group='+group+'&snapshot_id='+snapshotData+projectData,
		dataType: 'json',
		cache: false,
		success: function (data, textStatus) {
			tokens = filterSameLineUsage(data.tokens);
			showUsage(tokens, 0, snapshotData, projectData);
		},
		error: ajaxError
	});
}

function filterSameLineUsage(tokens){
	var lasttoken = null;
	var result = [];
	for (idx=0; idx<tokens.length; idx++){
		var token = tokens[idx];
		if ((lasttoken!=null)&&(token.location.hash==lasttoken.location.hash)&&(token.location.line==lasttoken.location.line))
			continue;
		lasttoken = token;
		result.push(token);
	}
	return result;
}

function makeLink(text){
	//It's dangerous to go alone, take this
	return $("<a href='javascript:;'>"+text+"</a>");
}

function showUsage(tokens, offset, snapshotData, projectData){

	var html = $('<div style="white-space: nowrap; width:100%; min-height:100%;"/>');
	var byFile={};
	var lastKey=undefined;
	var collector;
	var totalFileNum=0;

	//we now group usage by files since we will show em grouped by files later
	for (var idx=0; idx<tokens.length; idx++){
		var token = tokens[idx];
		var key=token.location.file_path;
		if (key!=lastKey){
			var collector=byFile[key];
			if (collector==undefined){
				totalFileNum++;
				collector=[];
				byFile[key]=collector;
			}
			lastKey=key;
		}
		collector.push(token);
	}

	var itemsPerPage = 15;
	$('#usagesContainer').html('');
	$('#usagesContainer').append(html);
	
	//this is extracted into function because we call it two times: first time to immediately show first entries, second time to render remaining entries when user asks for it
	var renderer = function(start, showAll){
		if (showAll==undefined)
			showAll = false;
		var fileIdx=0;
		for (var file in byFile) {
			if (byFile.hasOwnProperty(file)) {
				if (showAll||((fileIdx>=start)&&(fileIdx<(start+itemsPerPage)))){
					var list = byFile[file];
					var token = list[0];
					var fileDiv = $('<div/>');
					html.append(fileDiv);
					var fileNameDiv = $('<div/>'); 
					fileNameDiv.text(token.location.short_name.split(':')[0]);
					fileNameDiv.attr('title', token.location.file_path);
					fileNameDiv.css('display', 'inline');
					fileDiv.append(fileNameDiv);
					var filler = function(navigable){
						var llist=list; //I do not know js
						var small=$('<div style="display:inline;"></div>');
						var lineRenderer=function(ltoken){
							if (navigable){
								var link=$("<a id='usage_"+idx+"' href='javascript:;'>"+ltoken.location.line+"</a>");
								link.click([ltoken],function(data){loadSource(data.data[0].location, projectData, snapshotData);});
								small.append(link);
							} else {
								small.append(ltoken.location.line);
							}
							small.append(' ');
						}
						lineRenderer(token);
						if (llist.length>1){
							var expand = makeLink('Show more...');
							expand.click(function(){
								small.html('');
								llist.forEach(function(token){
									lineRenderer(token);
								});
							});
							small.append(expand);
						}
						fileDiv.append(small);
					}
					if ((token.location.hash==null)||(token.location.hash=='null')){
						fileDiv.append(' [missing source] : ');
						filler(false);
						fileDiv.css('color', '#aaaaaa');
					} else {
						fileDiv.append(' : ');
						filler(true);
/*						list.forEach(function(token){
						});*/
					}
				} else if (fileIdx==(start+itemsPerPage)){
/*					var link=$("<a id='usage_moar' href='javascript:;'>Load remaining usage (may take some time)</a>");
					link.click(function(){$('#usage_moar').hide();renderer(itemsPerPage,totalFileNum-itemsPerPage);});
					html.append(link);*/
					break;
				}
				fileIdx++;
			}
		}
	}
	
	var renderPage = function (userPage, showAll){
		if (showAll==undefined)
			showAll = false;
		var page=userPage-1;
		html.html('');
		
		if (totalFileNum<=itemsPerPage)
			showAll = true;
		
		if (!showAll){
			var usageNavBar = $("<div/>");
			html.append(usageNavBar);
			html.append('<br/>');
			var totalPages = Math.ceil(totalFileNum/itemsPerPage);
			if (userPage!=1){
				var linkToFirstPage = makeLink('|&lt;&lt;');
				linkToFirstPage.click(function(){renderPage(1)});
				usageNavBar.append(linkToFirstPage);
				usageNavBar.append('&nbsp;');
				
				var linkToPrevPage = makeLink('&lt;');
				linkToPrevPage.click(function(){renderPage(userPage-1)});
				usageNavBar.append(linkToPrevPage);
				usageNavBar.append('&nbsp;');
			} else {
				usageNavBar.append("<font color='#888888'>|&lt;&lt;&nbsp;&lt;&nbsp;</font>");
			}
			
			usageNavBar.append('page ' + userPage + '/' + totalPages);

			if (userPage!=totalPages){
				usageNavBar.append('&nbsp;');
				var linkToNextPage = makeLink('&gt;');
				linkToNextPage.click(function(){renderPage(userPage+1)});
				usageNavBar.append(linkToNextPage);
			
				usageNavBar.append('&nbsp;');
				var linkToLastPage = makeLink('&gt;&gt;|');
				linkToLastPage.click(function(){renderPage(totalPages)});
				usageNavBar.append(linkToLastPage);
			} else {
				usageNavBar.append("<font color='#888888'>&nbsp;&gt;&nbsp;&gt;&gt;|</font>");
			}
			var showAllLink = makeLink('[all '+totalFileNum+' files]');
			showAllLink.click(function(){renderPage(1, true)});
			usageNavBar.append('&nbsp;');
			usageNavBar.append(showAllLink);
		}
		
		renderer(page*itemsPerPage, showAll);
	}
	
	renderPage(1);
}

var lastTreeBuild=undefined;
function refreshFileTree() {
	var proj = currentSvaceProject;
	if (lastTreeBuild==makeSnapshotIDParm(proj))
		return;
	lastTreeBuild=makeSnapshotIDParm(proj);
	$.ajax( {
		url:'/history/api?act=get_file_tree&snapshot_id='+lastTreeBuild+getProjectParameter(proj),
		dataType: 'json',
		cache: false,
		async: false,
		success: function (data, textStatus) {
			drawFileTree(data.root);
		},
		error: ajaxError
	});
}

function drawFileTree(data) {
	directoriesIDs={};
	$('#fileFilters').html("Loading...");
	setTimeout(function(){
		$('#fileFilters').html(generateFileTree(data,0));
	}, 1);
	
}

var directoriesIDs={};
function generateFileTree(data,level) {
	var directories="";
	var files="";
	var prefix = Array(level+1).join("|");
	for (var idx=0; idx<data.children.length; idx++){
		child=data.children[idx];
		if (child.type=='dir'){
			normalDirID = MD5(child.dir_id);
			directoriesIDs[normalDirID]=child.dir_id;
			directories+=
				prefix
				+'<input type="checkbox" id="check_'+normalDirID+'" onclick="dirClicked(\''+normalDirID+'\', true)", svace_type="d">'
				+'<a href="javascript:;" onclick="$(\'#dir_'+normalDirID+'\').toggle()">'+generateFileTreeNodeLabel(child)+'</a><br/>\n'
				+'<div id="dir_'+normalDirID+'" style="display:none;">'+generateFileTree(child, level+1)+'</div>';
		} else {
			if (child.file_id==null)
				checkBoxCode='<input type="checkbox" disabled="disabled" svace_type="u">'
			else
				checkBoxCode='<input type="checkbox" id="check_'+child.file_id+'" onclick="fileClicked(\''+child.file_id+'\', true)" svace_type="f">';
			files+=
				prefix
				+checkBoxCode
				+generateFileTreeNodeLabel(child)
				+"<br/>";
		}
	}
	return directories+files;
}

function generateFileTreeNodeLabel(node){
	var result='';
	if (node.type=='dir')
		result+=escapeHTML(node.dir_name)
	else
		result+=escapeHTML(node.file_name)
	result+="<font color='#ff0000'>("+node.warnings+")</font>";
	return result;
}

function dirClicked(dirID, interactive){
	var element=$('#check_'+dirID);
	var value = element.prop('checked');
	if (interactive){
		$('#dir_'+dirID+' input').each(function(){
			var thisID=$(this).attr('id');
			if (thisID!=undefined){
				$(this).prop('checked', value);
				fileClicked(thisID.substring(6), interactive);
			}
		});
	}
	fileClicked(dirID, interactive);
}

function fileClicked(fileID, interactive){
	var element=$('#check_'+fileID);
	var value = element.prop('checked');
	var parentID=element.parent().attr('id');
	if (parentID!='fileFilters'){
		parentID = parentID.substring(4);
		if (value){
			collector = true;
			$('#dir_'+parentID+' > input').each(function(){
				if ((!$(this).prop('checked'))&&($(this).attr('id')!=undefined))
					collector = false;
			});
			if (collector){
				$('#check_'+parentID).prop('checked', true);
				dirClicked(parentID, false);
			}
		} else {
			$('#check_'+parentID).prop('checked', false);
			dirClicked(parentID, false);
		}
	}
	return element;
}
